!
!  Copyright (C) 2012, Northwestern University and Argonne National
!  Laboratory
!  See COPYRIGHT notice in top-level directory.
!
! $Id$

!    This example is the Fortran 77 version of nonblocking_write.c
!    It creates a netcdf file in CD-5 format and writes a number of
!    3D integer non-record variables.
!    Usage: (for example)
!    To compile:
!        mpif77 -O2 nonblocking_write.f -o nonblocking_write -lpnetcdf
!    To run (for example):
!        mpiexec -n 32 ./nonblocking_write /pvfs2/wkliao/testfile.nc 10
!    The size of each local array is len x len x len. Each non-record
!    variable is of size len*len*len * nprocs * sizeof(int)
!    All variables are partitioned among all processes in a 3D
!    block-block-block fashion.
!

      subroutine check(err, message)
          implicit none
          include "mpif.h"
          include "pnetcdf.inc"
          integer err
          character message*(*)

          ! It is a good idea to check returned value for possible error
          if (err .NE. NF_NOERR) then
              write(6,*) message//' '//nfmpi_strerror(err)
              call MPI_Abort(MPI_COMM_WORLD, -1, err)
          end if
      end

      program main
          implicit none
          include "mpif.h"
          include "pnetcdf.inc"

          character*256 filename, cmd, str
          integer i, j, cmode, err, ierr, info
          integer rank, nprocs, ncid, var_size, buf_size
          integer :: dimids(1), varids(1)
          integer, allocatable :: req(:), st(:)
          integer*8, allocatable:: starts(:), counts(:)
          integer, allocatable :: local_starts(:)
          integer*8 :: length
          real*8, allocatable  :: buf1d(:)
          integer :: color1, color2, comm1, comm2, grank, group_size, readi

          call MPI_Init(err)
          call MPI_Comm_rank(MPI_COMM_WORLD, rank, err)
          call MPI_Comm_size(MPI_COMM_WORLD, nprocs, err)

          ! take filename from command-line argument if there is any
          if (rank .EQ. 0) then
              filename = "testfile.nc"
          endif

          call MPI_Bcast(filename, 256, MPI_CHARACTER, 0, MPI_COMM_WORLD, err)

          group_size = 10
          color1 = rank/group_size
          color2 = mod(rank, group_size)
          call mpi_comm_split(MPI_COMM_WORLD,color1,rank,comm1,err)
          call mpi_comm_split(MPI_COMM_WORLD,color2,rank,comm2,err)
          call mpi_comm_rank(comm1, grank, err)

          ! read start & count
          if(grank .eq. 0) then
             readi = rank/group_size*8
             if(readi .lt. 10) then
               write(str,'(I1)') readi
             else
               write(str,'(I2)') readi
             end if
             open(11, file="start/Start"//str)
             read(11, *) var_size

             allocate(starts(var_size))
             allocate(counts(var_size))
             do i = 1, var_size
                read(11, *) starts(i)
             end do
             close(11)
             open(22, file="count/Count"//str)
             do i = 1, var_size
                read(22, *) counts(i)
             end do
             close(22)
             allocate(local_starts(var_size))
             local_starts(1) = 1
             do i = 2, var_size
                local_starts(i) = local_starts(i-1) + counts(i-1)
             end do
             buf_size = sum(counts)
             allocate(buf1d(buf_size))
             allocate(req(var_size))
             allocate(st(var_size))
          end if

          ! set an MPI-IO hint to disable file offset alignment for
          ! fixed-size variables
          !call MPI_Info_create(info, err)
          !call MPI_Info_set(info, "nc_var_align_size", "1", err)

          ! create file, truncate it if exists
          info = MPI_INFO_NULL
          cmode = IOR(NF_CLOBBER, NF_64BIT_DATA)
          if(grank .eq. 0) then
            err = nfmpi_create(comm2, filename, cmode, info, ncid)
            call check(err, 'In nfmpi_create: ')

            !call MPI_Info_free(info, err)

            ! define dimensions
            length = 655362
            do i=1, 1
               write(str,'(I1)') i-1
               err = nfmpi_def_dim(ncid, "x"//str, length, dimids(i))
               call check(err, 'In nfmpi_def_dim x'//str)
            enddo

            !define variables
            do i=1, 1
               write(str,'(I1)') i-1
               err = nfmpi_def_var(ncid, "var"//str, NF_DOUBLE, 1, dimids, varids(i))
               call check(err, 'In nfmpi_def_var var'//str)
            enddo

            ! do not forget to exit define mode
            err = nfmpi_enddef(ncid)
            call check(err, 'In nfmpi_enddef: ')

            ! write one variable at a time using iput
            do i = 1, buf_size
               buf1d(i) = rank
            end do

            do j = 1, var_size
               err = nfmpi_iput_vara_double(ncid, varids(1), starts(j), &
                     counts(j), buf1d(local_starts(j):local_starts(j)+counts(j) ), req(j))
            end do

            print*, "ffffffffff", rank
            ! wait for the nonblocking I/O to complete
            err = nfmpi_wait_all(ncid, var_size, req, st)
            print*, "hhhhhhhhhh", rank
            call check(err, 'In nfmpi_wait_all')

            ! check the status of each nonblocking request
            do i=1, 1
               write(str,'(I1)') i-1
               call check(st(i), 'In nfmpi_wait_all req '//str)
            enddo

            ! close the file
            err = nfmpi_close(ncid)
            call check(err, 'In nfmpi_close')
          end if

          ! check if there is any PnetCDF internal malloc residue
 998      format(A,I13,A)

 999      call MPI_Finalize(err)
          if(grank .eq. 0) then
            deallocate(starts)
            deallocate(counts)
            deallocate(local_starts)
            deallocate(buf1d)
            deallocate(req)
            deallocate(st)
          end if
          ! call EXIT(0) ! EXIT() is a GNU extension

      end

